""" libmup.py - M.A.R-T>Y. Micro Protocol (MuP) is an endpoint-based request/response RPC
 * protocol for small-ish interlinked systems.

 * NOTE: This python version is a somewhat close logic port of the C version.
 * Although some of the 'pythonic' stuff makes protocol work somewhat yucky...

 * See the C version for comments about logic. WARNING: This py version doesn't have built
 * in config stuff like bring your own crc, but if you've got the aptitude to enable
 * that feature you can mod it into here on your own :P

 * The protocol spec can be found at https://hoff.industries/release-mup.html
 * Last updated 06/08/2024

 * please report bugs to theworldisyours@mailbox.org

 * This library licensed with BSD 3-Clause:

Copyright 2023-2024 hoff.industries

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."""

from enum import Enum

MUP_LIB_VERSION = (0, 1, 4)
MUP_PROTO_VERSION = b'\x01'

MUP_AUTHOR = "hoff.industries"
MARTY_FLAVOUR_TEXT = "M.A.R-T>Y."

MUP_MAX_SIZE = 2048
MUP_HEADER_SIZE_W_CRC = 9
MUP_HEADER_SIZE_WO_CRC = 5

MUP_PAYLOAD_TERM = b'\xFE'
MUP_PACKET_TERM = b'\xFF'

MUP_SEND_PAYL_ENDPOINT_IDX = 0

MUP_MAX_PAYLOAD_SIZE = MUP_MAX_SIZE - MUP_HEADER_SIZE_W_CRC - 3

MUP_HEADER_IDX = {

	"TYPE": 0,
	"ARG": 1,
	"VERSION": 2,
	"SIZE": 3,
	"CRC": 5
}

MUP_TYPES = {

	"UNSET": b'\x00',
	"INIT": b'\x01',
	"BAD": b'\x02',
	"CONTERM": b'\x03',
	"SONAR": b'\x04',
	"SEND": b'\x05',
	"API": b'\x06'
}

MUP_TYPES_REVERSE = {
	b'\x00': "UNSET",
	b'\x01': "INIT",
	b'\x02': "BAD",
	b'\x03': "CONTERM",
	b'\x04': "SONAR",
	b'\x05': "SEND",
	b'\x06': "API"
}

MUP_PAYLOAD_COUNTS = {

	"UNSET": 0,
	"INIT": 1,
	"BAD": 0,
	"CONTERM": 0,
	"SONAR": 0,
	"SEND": 1,
	"API": 0
}

MUP_ARGS = {

	"UNSET": {

		"UNSET":	b'\x00'
	},

	"INIT": {

		"INIT":			b'\x01',
		"ACCEPT":		b'\x02',
		"REJECT":		b'\x03'
	},

	"BAD": {

		"TYPE":			b'\x01',
		"PACK_LEN":		b'\x02',
		"ARG":			b'\x03',
		"PAYL":			b'\x04',
		"PAYL_COUNT":	b'\x05',
		"VERSION":		b'\x0A',
		"CRC":			b'\x0B',
		"INTERN_ERR":	b'\x0C'
	},

	"CONTERM": {

		"CLEAN":		b'\x01',
		"SPAM":			b'\x02',
		"MANY_BAD":		b'\x03'
	},

	"SONAR": {

		"PING":			b'\x00',
		"PONG":			b'\x01'
	},

	"SEND": {

		"DIRECT":		b'\x01'
	},

	"API": {

		"SUCCESS":		b'\x01',
		"ERR_BUSY":		b'\x03',
		"ERR_RANGE":	b'\x04',
		"ERR_ENDPOINT":	b'\x05',
		"ERR_VALUE":	b'\x06',
		"ERR_COUNT":	b'\x07',
		"ERR_EXEC":		b'\x0A'
	}
}

# CRC32 table and routine sourced from
# https://web.mit.edu/freebsd/head/sys/libkern/crc32.c
# it allows unrestricted use

# CRC-32/ISO-HDLC poly 0x04C11DB7
CRC32_TAB = [

	0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
	0xe963a535, 0x9e6495a3,	0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
	0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
	0xf3b97148, 0x84be41de,	0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
	0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,	0x14015c4f, 0x63066cd9,
	0xfa0f3d63, 0x8d080df5,	0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
	0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,	0x35b5a8fa, 0x42b2986c,
	0xdbbbc9d6, 0xacbcf940,	0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
	0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
	0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
	0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,	0x76dc4190, 0x01db7106,
	0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
	0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
	0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
	0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
	0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
	0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
	0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
	0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
	0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
	0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
	0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
	0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
	0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
	0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
	0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
	0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
	0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
	0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
	0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
	0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
	0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
	0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
	0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
	0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
	0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
	0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
	0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
	0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
	0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
	0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
	0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
	0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
]

class MUPError(Enum):
	"""errors the lib could encounter"""

	MUP_ERR_NONE =				0

	MUP_ERR_OOB =				1
	MUP_ERR_INTERNAL =			2

	MUP_ERR_BAD_VERSION =		3
	MUP_ERR_BAD_CRC =			4
	MUP_ERR_BAD_TYPE =			5
	MUP_ERR_BAD_ARG =			6
	MUP_ERR_BAD_PACK_LEN =		7
	MUP_ERR_BAD_PAYL_COUNT =	8
	MUP_ERR_BAD_PAYL_VALUE =	9

	# python specific errs are given
	# big numbers to keep the rest in sync
	# with the C library
	MUP_ERR_BAD_PACK_TYPE =		100

MUP_ERR_ARG_LOOKUP = {

	MUPError.MUP_ERR_NONE:				MUP_ARGS["BAD"]["INTERN_ERR"],
	MUPError.MUP_ERR_OOB:				MUP_ARGS["BAD"]["PACK_LEN"],
	MUPError.MUP_ERR_BAD_CRC:			MUP_ARGS["BAD"]["CRC"],
	MUPError.MUP_ERR_BAD_VERSION:		MUP_ARGS["BAD"]["VERSION"],
	MUPError.MUP_ERR_BAD_TYPE:			MUP_ARGS["BAD"]["TYPE"],
	MUPError.MUP_ERR_BAD_ARG:			MUP_ARGS["BAD"]["ARG"],
	MUPError.MUP_ERR_BAD_PACK_LEN:		MUP_ARGS["BAD"]["PACK_LEN"],
	MUPError.MUP_ERR_BAD_PAYL_COUNT:	MUP_ARGS["BAD"]["PAYL_COUNT"],
	MUPError.MUP_ERR_BAD_PAYL_VALUE:	MUP_ARGS["BAD"]["PAYL"]
}

class MUPEvent(Enum):
	"""events handleable by your application"""

	MUP_EVT_NONE =			0
	MUP_EVT_INIT =			1
	MUP_EVT_INTERN_ERR =	2
	MUP_EVT_API_ERR =		3
	MUP_EVT_API_SUCCESS =	4
	MUP_EVT_RECV =			5
	MUP_EVT_MY_BAD =		6
	MUP_EVT_YOUR_BAD =		7
	MUP_EVT_CORRUPTED =		8
	MUP_EVT_SONAR_PING =	9
	MUP_EVT_SONAR_PONG =	10
	MUP_EVT_CONTERM =		11

def bytelist(input_bytes: bytearray) -> list:
	"""convert a bytearray which for some reason is all ints into
		an actual byte type"""

	return [int.to_bytes(byte, length=1, byteorder="big") \
		for byte in list(input_bytes)]

def mup_crc32(data: bytearray, size: int=-1) -> int:
	"""calc CRC32 based on the table"""

	if size == -1:
		size = len(data)

	crc = 0xFFFFFFFF

	for i in range(0, size):
		crc = CRC32_TAB[(crc ^ data[i]) & 0xFF] ^ (crc >> 8)

	return crc ^ 0xFFFFFFFF

class MUStatus:
	"""internal library and event status"""

	def __init__(self):

		self._size = 0
		self._error = MUPError.MUP_ERR_NONE
		self._event = MUPEvent.MUP_EVT_NONE
		self._do_reply = False
		self._more_waiting = False

	def mup_status_factory(self):
		"""reset status to defaults"""
		self.__init__()

	def get_size(self) -> int:
		"""self-explanatory"""
		return self._size

	def set_size(self, size: int):
		"""self-explanatory"""
		self._size = size

	def get_error(self) -> MUPError:
		"""self-explanatory"""
		return self._error

	def set_error(self, error: MUPError):
		"""self-explanatory"""
		self._error = error

	def get_event(self) -> MUPEvent:
		"""self-explanatory"""
		return self._event

	def set_event(self, event: MUPEvent):
		"""self-explanatory"""
		self._event = event

	def get_reply(self) -> bool:
		"""self-explanatory"""
		return self._do_reply

	def set_reply(self, reply: bool):
		"""self-explanatory"""
		self._do_reply = reply

	def get_more_waiting(self) -> bool:
		"""are there extra packets in the buffer?"""
		return self._more_waiting

	def set_more_waiting(self, more_waiting: bool):
		"""self-explanatory"""
		self._more_waiting = more_waiting

class MUEndpoint:
	"""niceties for defining an endpoint"""

	def __init__(self, endpoint: int, name: str,
			minimum: int=0, maximum: int=0):

		self._endpoint = endpoint
		self._name = name
		self._minimum = minimum
		self._maximum = maximum

	def get_endpoint(self) -> int:
		"""get the index of the endpoint in the endpoint list"""
		return self._endpoint

	def set_endpoint(self, endpoint: int):
		"""set the index of the endpoint in the endpoint list"""
		self._endpoint = endpoint

	def get_name(self) -> str:
		"""self-explanatory"""
		return self._name

	def set_name(self, name: str):
		"""self-explanatory"""
		self._name = name

	def get_min(self) -> int:
		"""get the minimum number of payloads for the endpoint"""
		return self._minimum

	def set_min(self, minimum: int):
		"""set the minimum number of payloads for the endpoint"""
		self._minimum = minimum

	def get_max(self) -> int:
		"""get the max number of payloads for the endpoint"""
		return self._maximum

	def set_max(self, maximum: int):
		"""set the max number of payloads for the endpoint"""
		self._maximum = maximum

class MUPacket:
	"""represents a packet on and off the wire"""

	def __init__(self):

		self._type = MUP_TYPES["UNSET"]
		self._arg = MUP_ARGS["UNSET"]["UNSET"]
		self._version = MUP_PROTO_VERSION
		self._size = 0
		self._crc32 = 0x00000000

		self._payloads = []
		self._data = bytearray()

		self._packet_index = 0

	def mup_packet_factory(self, zero: bool):
		"""reset packet"""

		self._version = MUP_PROTO_VERSION
		self._type = MUP_TYPES["UNSET"]
		self._arg = MUP_ARGS["UNSET"]["UNSET"]
		self._payloads = []
		self._size = 0
		self._crc32 = 0x00000000

		if zero:
			self._data = bytearray()

	def get_type(self) -> bytes:
		"""self-explanatory"""
		return self._type

	def set_type(self, mtype: bytes):
		"""self-explanatory"""
		self._type = mtype

	def get_arg(self) -> bytes:
		"""self-explanatory"""
		return self._arg

	def set_arg(self, arg: bytes):
		"""self-explanatory"""
		self._arg = arg

	def get_version(self) -> bytes:
		"""self-explanatory"""
		return self._version

	def set_version(self, version: bytes):
		"""self-explanatory"""
		self._version = version

	def get_size(self) -> int:
		"""self-explanatory"""
		return self._size

	def set_size(self, size: int):
		"""self-explanatory"""
		self._size = size

	def get_payloads(self) -> list:
		"""self-explanatory"""
		return self._payloads

	def set_payloads(self, payloads: list):
		"""self-explanatory"""
		self._payloads = payloads

	def get_data(self) -> bytearray:
		"""self-explanatory"""
		return self._data

	def set_data(self, data: bytearray):
		"""self-explanatory"""
		self._data = data

	def get_payload_count(self) -> int:
		"""self-explanatory"""
		return len(self._payloads)

	def to_string(self):
		"""Get a string of me for debug purposes"""
		return f"{self._type} {self._arg} {self._version} {self._size} {self._crc32}"

	def print_data(self):
		"""Print data in nice format"""
		for byte in self._data:
			print(hex(byte), end=" ")
		print("")

	def mup_serialize(self, status: MUStatus):
		"""Put data on the wire using self._payloads as payloads if applicable"""

		status.mup_status_factory()

		if self._type == MUP_TYPES["UNSET"]:

			status.set_error(MUPError.MUP_ERR_BAD_TYPE)
			return

		try:
			MUP_TYPES_REVERSE[self._type]
		except KeyError:

			status.set_error(MUPError.MUP_ERR_BAD_TYPE)
			return

		valid_payload_count = MUP_PAYLOAD_COUNTS[MUP_TYPES_REVERSE[self._type]]

		self._data = bytearray()

		self._data.extend(self._type)
		self._data.extend(self._arg)
		self._data.extend(self._version)

		if valid_payload_count == 0:

			status.set_size(MUP_HEADER_SIZE_WO_CRC + 1)
			self._data.extend(status.get_size().to_bytes(2, "big"))
			self._data.extend(MUP_PACKET_TERM)
			return

		# make space for size:
		self._data.extend(b'\x00\x00')
		# make space for CRC32:
		self._data.extend(b'\x00\x00\x00\x00')

		if self._type != MUP_TYPES["SEND"] and valid_payload_count != len(self._payloads):

			status.set_error(MUPError.MUP_ERR_BAD_PAYL_COUNT)
			return

		for payload in self._payloads:

			# not necessarily a fan of how this looks but
			# it should at least give flexibility with types

			if isinstance(payload, str):

				self._data.extend(payload.encode("ASCII"))
				self._data.extend(MUP_PAYLOAD_TERM)

			elif isinstance(payload, bytes):

				payload_bytearray = bytearray(payload)
				payload_bytelist = bytelist(payload_bytearray)

				self._data.extend(payload_bytearray)

				if payload_bytelist[-1] != MUP_PAYLOAD_TERM:
					self._data.extend(MUP_PAYLOAD_TERM)

			elif isinstance(payload, bytearray):

				payload_bytelist = bytelist(payload_bytearray)

				self._data.extend(payload)

				if payload_bytelist[-1] != MUP_PAYLOAD_TERM:
					self._data.extend(MUP_PAYLOAD_TERM)

			elif isinstance(payload, list):

				payload_bytearray = bytearray()
				for value in payload:

					if isinstance(value, int):
						payload_bytearray.extend(value.to_bytes(1, "big"))

					elif isinstance(value, bytes):
						payload_bytearray.extend(value)

					else:
						status.set_error(MUPError.MUP_ERR_BAD_PAYL_VALUE)
						return

				if payload[-1] != MUP_PAYLOAD_TERM:
					payload_bytearray.extend(MUP_PAYLOAD_TERM)

				self._data.extend(payload_bytearray)

			else:

				status.set_error(MUPError.MUP_ERR_BAD_PAYL_VALUE)
				return

		self._data.extend(MUP_PACKET_TERM)
		status.set_size(len(self._data))

		self._data[MUP_HEADER_IDX["SIZE"]] = status.get_size() >> 8
		self._data[MUP_HEADER_IDX["SIZE"] + 1] = status.get_size() & 0xFF

		self._crc32 = mup_crc32(self._data, status.get_size())

		self._data[MUP_HEADER_IDX["CRC"]] = (self._crc32 >> 24) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 1] = (self._crc32 >> 16) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 2] = (self._crc32 >> 8) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 3] = self._crc32 & 0xFF

	def mup_deserialize(self, status: MUStatus, all_strings: bool=False):
		"""Take data off the wire and put payloads into a nice list. Set all_strings to True
			to get every payload automatically converted to a string. """

		status.mup_status_factory()

		print(self._size, self._data[self._size:len(self._data)])

		if isinstance(self._data, bytes):
			packet_bytearray = bytearray(self._data[self._size:len(self._data)])
			self._data = packet_bytearray

		elif isinstance(self._data, bytearray):
			packet_bytearray = self._data[self._size:len(self._data)]
			self._data = packet_bytearray

		else:
			status.set_error(MUPError.MUP_ERR_BAD_PACK_TYPE)
			return

		if len(self._data) < MUP_HEADER_SIZE_WO_CRC:
			status.set_error(MUPError.MUP_ERR_BAD_PACK_LEN)
			return

		self.mup_packet_factory(False)

		packet_bytelist = bytelist(packet_bytearray)

		self._version = packet_bytelist[MUP_HEADER_IDX["VERSION"]]
		self._type = packet_bytelist[MUP_HEADER_IDX["TYPE"]]

		try:
			MUP_TYPES_REVERSE[self._type]
		except KeyError:

			status.set_error(MUPError.MUP_ERR_BAD_TYPE)
			return

		self._arg = packet_bytelist[MUP_HEADER_IDX["ARG"]]

		if not any(MUP_ARGS[MUP_TYPES_REVERSE[self._type]][arg] == self._arg \
				for arg in MUP_ARGS[MUP_TYPES_REVERSE[self._type]]):

			status.set_error(MUPError.MUP_ERR_BAD_ARG)
			return

		self._size = packet_bytearray[MUP_HEADER_IDX["SIZE"]] << 8 | \
			packet_bytearray[MUP_HEADER_IDX["SIZE"] + 1]

		if self._size > MUP_MAX_SIZE:

			status.set_error(MUPError.MUP_ERR_BAD_PACK_LEN)
			return

		valid_payload_count = MUP_PAYLOAD_COUNTS[MUP_TYPES_REVERSE[self._type]]

		if valid_payload_count == 0:
			if packet_bytelist[self.get_size() - 1] != MUP_PACKET_TERM:

				status.set_error(MUPError.MUP_ERR_BAD_PACK_LEN)
				return

			status.set_more_waiting(len(packet_bytelist) > self.get_size())

			return

		self._crc32 = int.from_bytes(self._data[\
			MUP_HEADER_IDX["CRC"]:(MUP_HEADER_IDX["CRC"] + 4)], "big")

		for i in range(0, 4):
			self._data[MUP_HEADER_IDX["CRC"] + i] = 0x00

		calced_crc = mup_crc32(self._data, self._size)

		if calced_crc != self._crc32:
			status.set_error(MUPError.MUP_ERR_BAD_CRC)
			return

		# put it back for completeness
		self._data[MUP_HEADER_IDX["CRC"]] = (self._crc32 >> 24) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 1] = (self._crc32 >> 16) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 2] = (self._crc32 >> 8) & 0xFF
		self._data[MUP_HEADER_IDX["CRC"] + 3] = self._crc32 & 0xFF

		self._payloads.append([])

		current_payload = 0

		# this is kinda yuck and not very efficient but
		# hey whaddya want from me it's a python program
		# you're not going to be efficient anyway!
		for i in range(MUP_HEADER_SIZE_W_CRC, len(packet_bytelist)):

			if packet_bytelist[i] == MUP_PACKET_TERM:
				break

			self._payloads[current_payload].append(packet_bytelist[i])

			if packet_bytelist[i] == MUP_PAYLOAD_TERM:

				self._payloads.append([])
				current_payload += 1

		if current_payload < valid_payload_count:

			status.set_error(MUPError.MUP_ERR_BAD_PAYL_COUNT)
			return

		if self._type != MUP_TYPES["SEND"] and \
				current_payload != valid_payload_count:

			status.set_error(MUPError.MUP_ERR_BAD_PAYL_COUNT)
			return

		self._payloads = self._payloads[0:-1]

		status.set_more_waiting(len(packet_bytelist) > self.get_size())

		if not all_strings:
			return

		# convert every payload bytelist to a string!

		for i in range(0, len(self._payloads)):

			self._payloads[i] = self._payloads[i][0:len(self._payloads[i]) - 1]
			self._payloads[i] = "".join(char.decode("ASCII") for char in self._payloads[i])

	def mup_construct_api_msg(self, arg: bytes, status: MUStatus):
		"""does what it says on the tin"""

		status.mup_status_factory()
		self.mup_packet_factory(False)

		self._type = MUP_TYPES["API"]
		self._arg = arg

		self.mup_serialize(status)

	def mup_construct_bad_msg(self, status: MUStatus):
		"""does what it says on the tin"""

		status.mup_status_factory()
		self.mup_packet_factory(False)

		self._type = MUP_TYPES["BAD"]

		try:
			self._arg = MUP_ERR_ARG_LOOKUP[status.get_error()]
		except KeyError:
			status.set_error(MUPError.MUP_ERR_BAD_ARG)
			return

		self.mup_serialize(status)

def mup_process_events(rx_packet: MUPacket, tx_packet: MUPacket, \
		status: MUStatus, do_deserial: bool, all_strings: bool=False):

	"""Process and yield events for the application to handle"""

	status.mup_status_factory()

	if do_deserial:

		rx_packet.mup_deserialize(status, all_strings=all_strings)
		if status.get_error() != MUPError.MUP_ERR_NONE:

			store_err = status.get_error()

			if status.get_error() == MUPError.MUP_ERR_BAD_CRC:
				status.set_event(MUPEvent.MUP_EVT_CORRUPTED)
			else:
				status.set_event(MUPEvent.MUP_EVT_YOUR_BAD)

			store_evt = status.get_event()

			tx_packet.mup_construct_bad_msg(status)

			status.set_error(store_err)
			status.set_event(store_evt)
			status.set_reply(True)
			return

	rx_type = rx_packet.get_type()

	if rx_type == MUP_TYPES["INIT"]:
		if rx_packet.get_arg() == MUP_ARGS["INIT"]["ACCEPT"]:

			status.set_event(MUPEvent.MUP_EVT_INIT)
			return

		status.set_event(MUPEvent.MUP_EVT_INTERN_ERR)

	elif rx_type == MUP_TYPES["BAD"]:

		status.set_event(MUPEvent.MUP_EVT_MY_BAD)

	elif rx_type == MUP_TYPES["CONTERM"]:

		status.set_event(MUPEvent.MUP_EVT_CONTERM)

	elif rx_type == MUP_TYPES["SONAR"]:

		if rx_packet.get_arg() == MUP_ARGS["SONAR"]["PONG"]:

			status.set_event(MUPEvent.MUP_EVT_SONAR_PONG)
			return

		tx_packet.mup_packet_factory(False)

		tx_packet.set_type(rx_type)
		tx_packet.set_arg(MUP_ARGS[MUP_TYPES_REVERSE[rx_type]]["PONG"])
		tx_packet.mup_serialize(status)

		status.set_event(MUPEvent.MUP_EVT_SONAR_PING)
		status.set_reply(True)

	elif rx_type == MUP_TYPES["SEND"]:

		status.set_event(MUPEvent.MUP_EVT_RECV)

	elif rx_type == MUP_TYPES["API"]:

		if rx_type == MUP_ARGS["API"]["SUCCESS"]:

			status.set_event(MUPEvent.MUP_EVT_API_SUCCESS)
			return

		status.set_event(MUPEvent.MUP_EVT_API_ERR)

	else:

		status.set_error(MUPError.MUP_ERR_INTERNAL)
		status.set_event(MUPEvent.MUP_EVT_INTERN_ERR)

		store_err = status.get_error()
		store_evt = status.get_event()

		tx_packet.mup_construct_api_msg(MUP_ARGS["BAD"]["INTERN_ERR"], status)

		status.set_error(store_err)
		status.set_event(store_evt)
		status.set_reply(True)
